/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.ui.annotation.editingsupport.MultiInstanceAnnotationTextEditingSupport;

/**
 * <p>
 * Base class for {@link IAnnotationValueProvider}s that support {@link IAnnotatedSpecification} for
 * which multiple instances can be dynamically created. For each model element, a single instance of
 * this {@link IAnnotatedSpecification} exists whose storage is implemented by a {@link EMap}
 * {@code <EString, V>}, where the key is used as the instance name, and V is an EClass depicting
 * the annotation type (e.g., {@code EString, EIntegerObject, ...}.
 * </p>
 * 
 * <p>
 * The underlying {@link EMap} must be declared in the {@link IAnnotatedSpecification}'s meta-model
 * as detailed here: <a href="http://wiki.eclipse.org/EMF/FAQ#How_do_I_create_a_Map_in_EMF.3F">
 * http://wiki.eclipse.org/EMF/FAQ#How_do_I_create_a_Map_in_EMF.3F</a>.
 * </p>
 * 
 * @author diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class DynamicInstanceAnnotationValueProviderBase<T extends IAnnotatedSpecification>
		extends SingleEStructuralFeatureValueProviderBase<T> {

	/**
	 * {@link EDataType} of the value type V of the underlying EMap<String, V> storing the
	 * annotation
	 * instances.
	 */
	EDataType valueDataType;

	/**
	 * {@link EFactory} to construct (in particular from a {@link String} representation instances
	 * of
	 * the value type V of the underlying EMap<String, V> storing the annotation instances.
	 */
	EFactory valueFactory;

	/**
	 * Constructs a multi-instance enabled {@link IAnnotationValueProvider} for a
	 * {@link IAnnotatedSpecification} that whose storage is implemented by a single
	 * {@link EStructuralFeature}.
	 */
	public DynamicInstanceAnnotationValueProviderBase(EClass annotatedSpecificationEClass,
			EFactory annotationFactory, EStructuralFeature structuralFeature,
			EFactory attributeFactory, EDataType valueDataType, EFactory valueFactory) {
		super(annotatedSpecificationEClass, annotationFactory, structuralFeature, attributeFactory);
		this.valueDataType = valueDataType;
		this.valueFactory = valueFactory;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <V> V getAnnotationValue(T spec, String instanceKey) {
		// Return value of current instance (entry map containing all instances of the annotation)
		return ((EMap<String, V>)getAnnotationValue(spec)).get(instanceKey);

	}

	/**
	 * Sets a value for a multi-instance {@link IAnnotatedSpecification} from a {@link String}
	 * representation of the input.
	 */
	@SuppressWarnings("unchecked")
	private <V> void
			setAnnotationValueFromString(String value, T specification, String instanceKey) {
		((EMap<String, V>)getAnnotationValue(specification)).put(instanceKey,
				(V)valueFactory.createFromString(valueDataType, value));
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <V> void setAnnotationValue(V value, T specification, String instanceKey) {
		if(value instanceof String) {
			setAnnotationValueFromString((String)value, specification, instanceKey);
		} else {
			((EMap<String, V>)getAnnotationValue(specification)).put(instanceKey, value);
		}

	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey) {
		setAnnotationValueFromString(value, specification, instanceKey);
	}

	/** {@inheritDoc} */
	@Override
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey) {

		return new MultiInstanceAnnotationTextEditingSupport(viewer, clazz, instanceKey);
	}

	/** {@inheritDoc} */
	@Override
	public List<String> getInstanceKeys(T specification) {
		// Return the (sorted) list of dynamic instances created so far.
		EMap<String, ?> kVMap = getAnnotationValue(specification);

		if(kVMap != null) {
			List<String> rval = new ArrayList<String>();
			rval.addAll(kVMap.keySet());
			Collections.sort(rval);
			return rval;
		}
		return Collections.emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean allowsDynamicAnnotationInstances() {
		return true;
	}
}
