/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Base class for {@link IAnnotationValueProvider}s that manages a single {@link EStructuralFeature}
 * contained by a {@link IAnnotatedSpecification}.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class SingleEStructuralFeatureValueProviderBase<T extends IAnnotatedSpecification>
		extends ValueProviderBase<T> {

	/**
	 * The structural feature of the {@link IAnnotatedSpecification} that is managed by this
	 * {@link IAnnotationValueProvider}.
	 */
	final protected EStructuralFeature structuralFeature;

	/**
	 * {@link EFactory} of used to create new attribute instances.
	 */
	final protected EFactory structuralFeatureFactory;

	/**
	 * String representation of default value used to initialize the primitive attribute of the
	 * underlying {@link IAnnotatedSpecification}. If it is {@code null}, {@link EDataType}'s
	 * default value will be used.
	 */
	final protected String structuralFeatureDefaultValue;

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a single {@link EStructuralFeature}.
	 */
	public SingleEStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EFactory annotationFactory, EStructuralFeature structuralFeature,
			EFactory structuralFeatureFactory) {
		this(annotatedSpecificationEClass, annotationFactory, structuralFeature,
				structuralFeatureFactory, null);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a single {@link EStructuralFeature}.
	 */
	public SingleEStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EFactory annotationFactory, EStructuralFeature structuralFeature,
			EFactory structuralFeatureFactory, final String structuralFeatureDefaultValue) {

		super(annotatedSpecificationEClass, annotationFactory);

		this.structuralFeature = structuralFeature;
		this.structuralFeatureFactory = structuralFeatureFactory;
		this.structuralFeatureDefaultValue = structuralFeatureDefaultValue;
	}

	/**
	 * Sets a value for a {@link IAnnotatedSpecification} from a {@link String} representation of
	 * the input.
	 */
	private <V> void setAnnotationValueFromString(String value, T specification) {
		specification.eSet(
				structuralFeature,
				structuralFeatureFactory.createFromString(
						((EAttribute)structuralFeature).getEAttributeType(), value));
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		if(value instanceof String) {
			setAnnotationValueFromString((String)value, specification);
		} else {
			specification.eSet(structuralFeature, value);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		if(structuralFeature instanceof EAttribute) {
			setAnnotationValueFromString(value, specification);
		} else {
			throw new Exception(
					"setAnnotationValue(String, T) is not supported / has not been implemented the annotation type " +
							structuralFeature.getEType().getName() +
							"(specification class: " +
							specification.getClass() + ").");
		}
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification) {
		return (U)specification.eGet(structuralFeature);
	}

	/**
	 * Creates a new instance of the {@link EStructuralFeature} (use to populate newly created
	 * annotation instances created in
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 */
	protected Object createStructuralFeatureInstance() throws Exception {
		if(structuralFeature instanceof EAttribute) {

			EDataType attributeEDataType = ((EAttribute)structuralFeature).getEAttributeType();

			if(structuralFeatureDefaultValue != null) {
				// Use default value if it has been specified
				return structuralFeatureFactory.createFromString(attributeEDataType,
						structuralFeatureDefaultValue);
			}

			// Return data types default value, if it is non-null.
			Object rval = attributeEDataType.getDefaultValue();

			if(rval != null) {
				return rval;
			}

			try {
				// Try to construct attribute from empty string
				rval = structuralFeatureFactory.createFromString(attributeEDataType, "");
			} catch(NumberFormatException e) {
				// Last try: initialize numbers with 0 (throws NumberFormatException if not
				// successful).
				rval = structuralFeatureFactory.createFromString(attributeEDataType, "0");
			}

			return rval;
		} else if(structuralFeature instanceof EReference) {
			EReference reference = (EReference)structuralFeature;
			if(reference.isContainment()) {
				// EClasses attached to contained EReferences are created lazily by the
				// corresponding generated code on the first access (i.e., also during get / "read")
				return null;
			}
			return structuralFeatureFactory.create(reference.getEReferenceType());
		} else {
			throw new Exception(
					"createStructuralFeatureInstance() is not supported / has not been implemented the annotation type " +
							structuralFeature.getEType().getName() + ".");
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void decorateAnnotationSpecification(T specification) throws Exception {
		// Create and set structural feature implementing the annotation
		Object structuralFeatureVal = createStructuralFeatureInstance();
		if(structuralFeatureVal != null) {
			specification.eSet(structuralFeature, structuralFeatureVal);
		}
	}

}
