/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import java.util.Collections;
import java.util.List;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.base.ui.annotation.editingsupport.TextEditingSupport;

/**
 * Base class for {@link IAnnotationValueProvider}s.
 * 
 * @author diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class ValueProviderBase<T extends IAnnotatedSpecification> implements
		IAnnotationValueProvider<T> {

	/**
	 * {@link EClass} of {@link IAnnotatedSpecification} that contains the respective structural
	 * feature and that is managed by this {@link IAnnotationValueProvider}.
	 */
	protected final EClass annotatedSpecificationEClass;
	/**
	 * {@link EFactory} of ECore model that contains {@link IAnnotatedSpecification} identified by
	 * {@code annotatedSpecificationEClass} (used to create new instances of the annotation).
	 */
	protected final EFactory annotationFactory;

	/**
	 * Constructs a new {@link IAnnotationValueProvider} for an annotation specified in a given
	 * {@link EClass}, and constructed using a given {@link EFactory}.
	 */
	public ValueProviderBase(EClass annotatedSpecificationEClass, EFactory annotationFactory) {
		this.annotatedSpecificationEClass = annotatedSpecificationEClass;
		this.annotationFactory = annotationFactory;
	}

	/** {@inheritDoc} */
	@Override
	public boolean setAnnotationName(String name, T specification) {
		// Renaming this annotation is not supported.
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		return getAnnotationValue(specification);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		setAnnotationValue(value, specification, null);
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		setAnnotationValue(value, specification, null);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception {
		throw new Exception(
				"setAnnotationValue(String) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		throw new Exception(
				"setAnnotationValue(U) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/** {@inheritDoc} */
	@Override
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey) {

		// Default to text editing cell.
		return new TextEditingSupport(viewer, clazz);
	}

	/**
	 * Creates the editing support without referring to the optional {@code instanceKey} (see
	 * {@link #createEditingSupport(ColumnViewer, Class, String)}).
	 */
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz) {

		return createEditingSupport(viewer, clazz, null);
	}

	/** {@inheritDoc} */
	@Override
	public List<Class<? extends EObject>> excludeModelElementTypeFromAnnotatedSpecification() {
		return null;
	}

	/**
	 * Add decorations to an annotation when its instance is newly created by
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 * 
	 * @param specification
	 *            annotation to be decorated
	 */
	protected void decorateAnnotationSpecification(T specification) throws Exception {
		// No decoration
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public T getAnnotatedSpecificationForModelElement(IModelElement element) throws Exception {
		// Return existing annotation
		for(IModelElementSpecification specification : element.getSpecifications()) {
			if(annotatedSpecificationEClass.equals(specification.eClass())) {
				return (T)specification;
			}
		}

		// Create annotation
		T specification = (T)annotationFactory.create(annotatedSpecificationEClass);
		decorateAnnotationSpecification(specification);

		// Hook specification to model element
		element.addSpecification(specification);

		return specification;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<T> getAnnotationClazz() {
		return (Class<T>)annotatedSpecificationEClass.getClass();
	}

	/** {@inheritDoc} */
	@Override
	public List<String> getInstanceKeys(T specification) {
		return Collections.emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean allowsDynamicAnnotationInstances() {
		return false;
	}

}
