/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view;

import java.util.Collection;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ElementNameEditingSupport;
import org.fortiss.tooling.base.ui.annotation.valueprovider.IAnnotationValueProvider;

/**
 * <p>
 * This class provides a generic annotation view whose methods are intended to be reused by child
 * classes. Inherited classes can avoid code duplication, e.g. the creation of the table and its
 * first row, using this class.
 * </p>
 * <p>
 * In order to support displaying and editing specific annotation types, the corresponding
 * {@link IAnnotationValueProvider#createEditingSupport(org.eclipse.jface.viewers.ColumnViewer, Class, String)}
 * factory method must return an appropriate {@link EditingSupport}.
 * </p>
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: D91763408246ECD3578DAD0E37BC4C41
 */
public class GenericAnnotationView extends AnnotationViewPartBase {

	/** The table viewer */
	protected TableViewer tableViewer;

	/**
	 * Character used to separte the annotation name from the instance name for multi-instance
	 * annotations.
	 */
	private final static String MULTIINSTANCE_NAME_SEPARATOR = ":";

	/**
	 * Data required to identify a column displaying a particular {@link IAnnotatedSpecification} in
	 * a column of the {@link GenericAnnotationView}. Used to sort columns (see
	 * {@link #compareTo(ColumnHandle)}).
	 */
	private class ColumnHandle implements Comparable<ColumnHandle> {
		/**
		 * {@link AnnotationEntry} from which a particular {@link IAnnotatedSpecification} is to be
		 * displayed in the column represented by this {@link ColumnHandle}.
		 */
		private AnnotationEntry entry;

		/**
		 * {@link IAnnotatedSpecification} from {@code entry} to be displayed in the column
		 * represented by this {@link ColumnHandle}.
		 */
		private IAnnotatedSpecification annotatedSpecification;

		/**
		 * Instance key identifying the particular instance of an {@link IAnnotatedSpecification}
		 * from {@code entry} to be displayed in the column represented by this {@link ColumnHandle}
		 * .
		 */
		private String instanceKey;

		/** Flag if this {@link ColumnHandle} represents a {@link CreateAnnotationInstanceColumn}. */
		private boolean isCreateColumn;

		/** Constructs a new {@link ColumnHandle}. */
		public ColumnHandle(AnnotationEntry entry, IAnnotatedSpecification annotatedSpecification,
				String instanceKey, boolean isCreateColumn) {

			this.entry = entry;
			this.annotatedSpecification = annotatedSpecification;
			this.instanceKey = instanceKey;
			this.isCreateColumn = isCreateColumn;
		}

		/**
		 * Returns the {@link AnnotationEntry} from which a particular
		 * {@link IAnnotatedSpecification} is to be
		 * displayed in the column represented by this {@link ColumnHandle}.
		 */
		public AnnotationEntry getEntry() {
			return entry;
		}

		/**
		 * Returns the {@link IAnnotatedSpecification} from {@code entry} to be displayed in the
		 * column represented by this {@link ColumnHandle}.
		 */
		public IAnnotatedSpecification getAnnotatedSpecification() {
			return annotatedSpecification;
		}

		/**
		 * Returns the instance key identifying the particular instance of an
		 * {@link IAnnotatedSpecification} from {@code entry} to be displayed in the column
		 * represented by this {@link ColumnHandle} .
		 */
		public String getInstanceKey() {
			return instanceKey;
		}

		/**
		 * Returns a flag if this {@link ColumnHandle} represents a
		 * {@link CreateAnnotationInstanceColumn}.
		 */
		public boolean isCreateColumn() {
			return isCreateColumn;
		}

		/** {@inheritDoc} */
		@Override
		public int compareTo(ColumnHandle other) {
			String columnName = getColumnName(this);
			String annotationName;

			int columnNameSeparatorIndex = columnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
			if(columnNameSeparatorIndex != -1) {
				annotationName = columnName.substring(0, columnNameSeparatorIndex);
			} else {
				annotationName = columnName;
			}

			String otherColumnName = getColumnName(other);
			String otherAnnotationName;
			int otherColumnNameSeparatorIndex =
					otherColumnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
			if(otherColumnNameSeparatorIndex != -1) {
				otherAnnotationName = otherColumnName.substring(0, otherColumnNameSeparatorIndex);
			} else {
				otherAnnotationName = otherColumnName;
			}

			// Place "create" columns (label: "<specification name>") after their instances
			// (which are labeled "<specification name>: <instance name>").
			if(annotationName.equals(otherAnnotationName)) {
				// Ensure uniqueness of "create" columns"
				if(isCreateColumn && other.isCreateColumn) {
					return 0;
				}
				if(isCreateColumn) {
					return 1;
				}
				if(other.isCreateColumn) {
					return -1;
				}
			}

			// Sort alphabetically by specification column label
			return columnName.compareTo(otherColumnName);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void update(Collection<AnnotationEntry> annotationEntries) {
		// Minimize flickering while updating the table
		tableViewer.getTable().setRedraw(false);

		/*
		 * Sorted set of {@link ColumnHandle} used to instantiate this {@link
		 * GenericAnnotationView}'s columns in the right order.
		 */
		Set<ColumnHandle> sortedColumnHandles = new TreeSet<ColumnHandle>();

		/*
		 * Dispose all columns of the table viewer, except the first column, which holds the names
		 * of the model elements
		 */
		while(tableViewer.getTable().getColumnCount() > 1) {
			tableViewer.getTable().getColumn(1).dispose();
		}

		// Aggregate required columns. Column order is defined by ColumnHandle.compareTo().
		for(AnnotationEntry entry : annotationEntries) {
			for(IAnnotatedSpecification spec : entry.getSpecificationsList()) {
				// Create new column for each instance of the current multi-instance annotation
				if((entry.allowsDynamicAnnotationInstances(spec.getClass())) ||
						!entry.getInstanceKeys(spec.getClass()).isEmpty()) {

					for(String instanceKey : entry.getInstanceKeys(spec.getClass())) {
						sortedColumnHandles.add(new ColumnHandle(entry, spec, instanceKey, false));
					}
				}

				// Create a new column for
				// - ordinary (single-instance) annotations
				// - "create" column for multi-instance annotations for which instances can be
				// created dynamically
				if(entry.getInstanceKeys(spec.getClass()).isEmpty() ||
						entry.allowsDynamicAnnotationInstances(spec.getClass())) {
					sortedColumnHandles.add(new ColumnHandle(entry, spec, null, entry
							.allowsDynamicAnnotationInstances(spec.getClass())));

				}
			}
		}

		// Instantiate columns
		for(ColumnHandle columnHandle : sortedColumnHandles) {
			createAnnotationColumn(columnHandle);
		}

		// Update the view
		tableViewer.getTable().setRedraw(true);
		if(!tableViewer.getTable().isDisposed()) {
			tableViewer.setInput(annotationEntries);
		}
		tableViewer.refresh();
	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		super.createPartControl(parent);
		Composite c = new Composite(parent, SWT.NULL);
		c.setLayout(new FillLayout(SWT.HORIZONTAL));

		tableViewer = new TableViewer(c, SWT.BORDER | SWT.FULL_SELECTION);
		Table table = tableViewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);

		createModelElementColumn();
	}

	/** Creates the (leading) column which displays the model elements */
	protected void createModelElementColumn() {

		final TableViewerColumn firstColumn = new TableViewerColumn(tableViewer, SWT.NONE);
		firstColumn.getColumn().setText("IModelElement");
		firstColumn.getColumn().setWidth(100);

		tableViewer.setContentProvider(new ArrayContentProvider());
		firstColumn.setLabelProvider(new ElementNameLabelProvider(this, firstColumn));
		firstColumn.setEditingSupport(new ElementNameEditingSupport(tableViewer));
	}

	/** Creates a column in tableViewer that displays the annotation given in spec */
	protected void createAnnotationColumn(ColumnHandle columnHandle) {

		TableColumn tableColumn;
		if(columnHandle.isCreateColumn()) {
			// Create column that contains a button for adding new instances of the
			// current annotation type.

			CreateAnnotationInstanceColumn column =
					new CreateAnnotationInstanceColumn(tableViewer, SWT.NONE,
							columnHandle.getAnnotatedSpecification(), columnHandle.getEntry());
			tableColumn = column.getColumn();

		} else {

			// Add new new column
			TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);
			tableColumn = column.getColumn();

			String specName = getColumnName(columnHandle);
			tableColumn.setText(specName);

			Class<? extends IAnnotatedSpecification> annotationClass =
					columnHandle.getAnnotatedSpecification().getClass();

			// Have the matching EditingSupport created for the current annotated specification
			// (work is delegated to respective IAnnotationValueProvider implementation)
			EditingSupport editingSupport =
					columnHandle.getEntry().createSpecificationEditElement(tableViewer,
							annotationClass, columnHandle.getInstanceKey());
			column.setEditingSupport(editingSupport);

			// Add column label provider, and set EditingSupport
			column.setLabelProvider(new AnnotationLabelProvider(annotationClass, this, columnHandle
					.getInstanceKey()));
		}

		tableColumn.setWidth(125);

	}

	/**
	 * Returns the name of a column for a given {@link IAnnotatedSpecification} represented by the
	 * given {@link ColumnHandle}.
	 */
	protected String getColumnName(ColumnHandle columnHandle) {
		// Determine column name
		String specName =
				columnHandle.getEntry().getSpecificationAnnotationName(
						columnHandle.getAnnotatedSpecification().getClass());
		if(specName == null || specName.isEmpty()) {
			specName = "<Unnamed Annotation>";
		}

		if(columnHandle.getInstanceKey() != null) {
			specName += MULTIINSTANCE_NAME_SEPARATOR + " " + columnHandle.getInstanceKey();
		}
		return specName;
	}

	/** {@inheritDoc} */
	@Override
	public void setFocus() {
		// nothing to do
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		tableViewer.getTable().dispose();
		super.dispose();
	}

}
